/*
   Copyright 1999 by CodeFab, Inc.   All rights reserved.

   This notice may not be removed from this source code.
   The use and distribution of this software is governed by the
   terms of the MiscKit license agreement.  Refer to the license
   document included with the MiscKit distribution for the terms.
   Details of the license agreement may be found at
   www.misckit.com or within the MiscKit distribution.

   There may be more information about this software on
   www.codefab.com.  Send queries to ask_codefab@codefab.com.
*/

// import APIs to NeXT provided frameworks
#import <Foundation/Foundation.h>

// import local framework's API
#import "CodeFabSupport.h"

// import API required througout this file's scope
#import "NSObject_CFRLoggingExtensions.h"

// the following NSObject category interface (no implementation) contains methods that are 
// only invoked after an appropriate test is performed to ensure that the target object 
// actually implements said method!


@interface NSObject (ResponseTestedMethods)
- application;
@end

@implementation NSObject (CFRLoggingExtensions)
/*"
 * The purpose of this category is to provide a standard, complete and
 * convenient set of methods for generating well formatted and information
 * logging messages.  API is also provided to generate various forms of
 * logging strings and to control the eventual destination of the logging
 * output.
"*/
#if 0
- (void)logWithFormat:(NSString *)aFormat, ...;
/*"
 * Cover for
 * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
"*/
{
    va_list ap;

    va_start(ap, aFormat);
    [NSObject logMethod: _cmd
               ofObject: self
             withFormat: aFormat
              arguments: ap
       showSelectorInfo: NO];
    va_end(ap);
}
#endif
+ (void)logWithFormat:(NSString *)aFormat, ...;
/*"
 * Cover for
   * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
"*/
{
    va_list ap;

    va_start(ap, aFormat);
    [NSObject logMethod: _cmd
               ofObject: self
             withFormat: aFormat
              arguments: ap
       showSelectorInfo: NO];
    va_end(ap);
}

#if 0
- (void)logWithFormat:(NSString *)aFormat arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logMethod: _cmd
                      ofObject: self
                    withFormat: aFormat
                     arguments: argList
              showSelectorInfo: NO];
}
#endif
+ (void)logWithFormat:(NSString *)aFormat arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logMethod: _cmd
                      ofObject: self
                    withFormat: aFormat
                     arguments: argList
              showSelectorInfo: NO];
}

- (void)logMethod: (SEL) aMethod withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;

    va_start(ap, aFormat);
    [NSObject logMethod: aMethod
               ofObject: self
             withFormat: aFormat
              arguments: ap
       showSelectorInfo: YES];
    va_end(ap);
}

+ (void)logMethod: (SEL) aMethod withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;

    va_start(ap, aFormat);
    [NSObject logMethod: aMethod
               ofObject: self
             withFormat: aFormat
              arguments: ap
       showSelectorInfo: YES];
    va_end(ap);
}

- (void)logMethod: (SEL) aMethod withFormat:(NSString *)aFormat arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logMethod: aMethod
                      ofObject: self
                    withFormat: aFormat
                     arguments: argList
              showSelectorInfo: YES];
}

+ (void)logMethod: (SEL) aMethod withFormat:(NSString *)aFormat arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logMethod: aMethod
                      ofObject: self
                    withFormat: aFormat
                     arguments: argList
              showSelectorInfo: YES];
}


- (void)logMethod:(SEL)aMethod
         ofObject:(id)anObject
       withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;

    va_start(ap, aFormat);
    [NSObject logMethod: aMethod
               ofObject: anObject
             withFormat: aFormat
              arguments: ap
       showSelectorInfo: YES];
    va_end(ap);
}

+ (void)logMethod:(SEL)aMethod
         ofObject:(id)anObject
       withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;

    va_start(ap, aFormat);
    [NSObject logMethod: aMethod
               ofObject: anObject
             withFormat: aFormat
              arguments: ap
       showSelectorInfo: YES];
    va_end(ap);
}

- (void)logMethod:(SEL)aMethod
         ofObject:(id)anObject
       withFormat:(NSString *)aFormat
        arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logMethod: aMethod
                      ofObject: anObject
                    withFormat: aFormat
                     arguments: argList
              showSelectorInfo: NO];
}

+ (void)logMethod:(SEL)aMethod
         ofObject:(id)anObject
       withFormat:(NSString *)aFormat
        arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logMethod: aMethod
                      ofObject: anObject
                    withFormat: aFormat
                     arguments: argList
              showSelectorInfo: NO];
}

- (NSString *)logStringWithFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;
    NSString *returnString;

    va_start(ap, aFormat);
    returnString = [NSObject logStringForMethod: _cmd
                                       ofObject: self
                                     withFormat: aFormat
                                      arguments: ap
                               showSelectorInfo: NO];
    va_end(ap);

    return returnString;
}

- (NSString *)logStringWithFormat:(NSString *)aFormat
                        arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    NSString *returnString;

    returnString = [NSObject logStringForMethod: _cmd
                                       ofObject: self
                                     withFormat: aFormat
                                      arguments: argList
                               showSelectorInfo: NO];

    return returnString;
}

- (NSString *)logStringForMethod: (SEL) aMethod
                      withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;
    NSString *returnString;


    va_start(ap, aFormat);
    returnString = [NSObject logStringForMethod: aMethod
                                       ofObject: self
                                     withFormat: aFormat
                                      arguments: ap
                               showSelectorInfo: YES];
    va_end(ap);

    return returnString;
}

- (NSString *)logStringForMethod: (SEL) aMethod
                      withFormat:(NSString *)aFormat
                       arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logStringForMethod: aMethod
                               ofObject: self
                             withFormat: aFormat
                              arguments: argList
                       showSelectorInfo: NO];
}

- (NSString *)logStringForMethod:(SEL)aMethod
                        ofObject:(id)anObject
                      withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;
    NSString *returnString;

    va_start(ap, aFormat);
    returnString = [NSObject logStringForMethod: aMethod
                                       ofObject: anObject
                                     withFormat: aFormat
                                      arguments: ap
                               showSelectorInfo: YES];
    va_end(ap);
    return returnString;
}

- (NSString *)logStringForMethod:(SEL)aMethod
                        ofObject:(id)anObject
                      withFormat:(NSString *)aFormat
                       arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logStringForMethod: aMethod
                               ofObject: anObject
                             withFormat: aFormat
                              arguments: argList
                       showSelectorInfo: YES];
}

+ (NSString *)logStringWithFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;
    NSString *returnString;

    va_start(ap, aFormat);
    returnString = [NSObject logStringForMethod: _cmd
                                       ofObject: self
                                     withFormat: aFormat
                                      arguments: ap
                               showSelectorInfo: NO];
    va_end(ap);

    return returnString;
}

+ (NSString *)logStringWithFormat:(NSString *)aFormat
                        arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    NSString *returnString;

    returnString = [NSObject logStringForMethod: _cmd
                                       ofObject: self
                                     withFormat: aFormat
                                      arguments: argList
                               showSelectorInfo: NO];

    return returnString;
}

+ (NSString *)logStringForMethod: (SEL) aMethod
                      withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;
    NSString *returnString;


    va_start(ap, aFormat);
    returnString = [NSObject logStringForMethod: aMethod
                                       ofObject: self
                                     withFormat: aFormat
                                      arguments: ap
                               showSelectorInfo: YES];
    va_end(ap);

    return returnString;
}

+ (NSString *)logStringForMethod: (SEL) aMethod
                      withFormat:(NSString *)aFormat
                       arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logStringForMethod: aMethod
                               ofObject: self
                             withFormat: aFormat
                              arguments: argList
                       showSelectorInfo: NO];
}

+ (NSString *)logStringForMethod:(SEL)aMethod
                        ofObject:(id)anObject
                      withFormat:(NSString *)aFormat, ...;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    va_list ap;
    NSString *returnString;

    va_start(ap, aFormat);
    returnString = [NSObject logStringForMethod: aMethod
                                       ofObject: anObject
                                     withFormat: aFormat
                                      arguments: ap
                               showSelectorInfo: YES];
    va_end(ap);
    return returnString;
}

+ (NSString *)logStringForMethod:(SEL)aMethod
                        ofObject:(id)anObject
                      withFormat:(NSString *)aFormat
                       arguments:(va_list)argList;
    /*"
     * Cover for
       * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
    "*/
{
    return [NSObject logStringForMethod: aMethod
                               ofObject: anObject
                             withFormat: aFormat
                              arguments: argList
                       showSelectorInfo: YES];
}

static BOOL loggingEnabled = YES;
+ (void)setLoggingEnabbled:(BOOL) aFlag;
/*"
 * Enables and disables the output of logging messages.  If aFlag is YES,
 * then logging messages will be displayed.
"*/
{
    loggingEnabled = aFlag;
}

+ (BOOL)loggingEnabled;
/*"
 * Returns YES if logging is enabled.  NO otherwise.
"*/
{
    return loggingEnabled;
}

static NSFileHandle *loggingFileHandle = nil;
static loggingFileHandleNotInitialized = NO;
+ (void)setLoggingFileHandle: (NSFileHandle *) aHandle;
/*"
 * Sets the file handle to which all logging messages are output to
 * aHandle.  Messages will only be written to aHandle if logging is
 * enabled.
"*/
{
    CFRSetVariableToObject(loggingFileHandle, aHandle);
}

+ (NSFileHandle *)loggingFileHandle;
/*"
 * Returns the file handle to which all logging output is written.
"*/
{
    return loggingFileHandle;
}

+ (void)reallyLogMethod:(SEL)aMethod
               ofObject:(id)anObject
             withFormat:(NSString *) aFormat
              arguments:(va_list)argList
       showSelectorInfo:(BOOL) aFlag;
    /*"
    * Writes the log entry to the logging file handle.  If NSFileHandle cannot
     * be made to handle whatever output features are desired, override this
     * method.  This method generates the string that is logged by invoking
     * #{+logStringForMethod:ofObject:withFormat:arguments:showSelectorInfo:}.
     *
     * Appends a newline ('\n') to the output.
     "*/
{
    NSString *outString;
    static NSData *newLineData = nil;

    if (loggingEnabled == NO)
        return;

    if (loggingFileHandleNotInitialized == NO) {
        [self setLoggingFileHandle: [NSFileHandle fileHandleWithStandardError]];
        loggingFileHandleNotInitialized = YES;
    }

    if (newLineData == nil)
        newLineData = [[@"\n" dataUsingEncoding: NSASCIIStringEncoding] retain];

    outString = [self logStringForMethod: aMethod
                                ofObject: anObject
                              withFormat: aFormat
                               arguments: argList
                        showSelectorInfo: aFlag];

    [loggingFileHandle writeData: [outString dataUsingEncoding: NSASCIIStringEncoding]];
    [loggingFileHandle writeData: newLineData];
}

+ (void)logMethod:(SEL)aMethod
         ofObject:(id)anObject
       withFormat:(NSString *)aFormat
        arguments:(va_list)argList
 showSelectorInfo:(BOOL) aFlag;
{
    [self reallyLogMethod:aMethod
                 ofObject:anObject
               withFormat:aFormat
                arguments:argList
         showSelectorInfo:aFlag];
}

+ (NSString *)logStringForMethod:(SEL)aMethod
                        ofObject:(id)anObject
                      withFormat:(NSString *)aFormat
                       arguments:(va_list)argList
                showSelectorInfo:(BOOL) aFlag;
/*"
 * Generates the actual logging message.  Override this method to customize
 * the format of the log messages.
 * 
 * If %aFlag is YES, then the logging message is prefixed with information
 * about aMethod and anObject.
"*/
{
    NSString *outString;

    // create formatted output
    outString = [[[NSString alloc] initWithFormat: aFormat
                                        arguments: argList] autorelease];

    if (aFlag) {
        NSString *methodLabel;

        methodLabel = [NSObject methodLabelForInstance: anObject andSelector: aMethod];

        outString = [NSString stringWithFormat: @"%@ %@",methodLabel, outString];
    }
    return outString;
}

+ (void) exit: (NSString *) msg;
/*"
 * Outputs msg to the logging file handle (ensuring that logging is
 * enabled).  Calls exit with a value of 1.
"*/
{
    loggingEnabled = NO;
    if (msg)
        [self logMethod: _cmd
              withFormat: @"Exited with msg:\n\t%@", msg];
    exit(1);
}

- (void) exit: (NSString *) msg;
/*"
 * Invokes the class method #{+exit:}.
"*/
{
    [[self class] exit: msg];
}

+ (NSString *) canonicalNameForClass: (id)theObject
/*"
 * Returns the canonical name for the class represented by theObject.
 * theObject may be a class or instance object.
 * 
 * If theObject represents an instance of a WebScript Scripted Class, the
 * returned name will indicate that the object is a scripted class and will
 * have the truncated name (instead of returning a name that includes the
 * relative path to the file that implements the class).
 * 
 * If theObject represents a subclass of WOComponent, the returned class
 * will contain the return value of the #{-name} method.
 * 
 * If theObject is a kind of WOSession, the returned string will include
 * the name of the class and the name of the WebObjects application within
 * which the session resides.
"*/
{
    Class oC;
    NSString *originalClassName;

    if (theObject == nil)
        return NSStringFromClass(nil);

    if ([theObject respondsToSelector: @selector(dfobjc_canonicalName)])
        return [theObject dfobjc_canonicalName];

    oC = [theObject class];

    if (oC == Nil)
        return nil;

    if ([oC respondsToSelector: @selector(dfobjc_canonicalName)])
        return [oC dfobjc_canonicalName];

    originalClassName = NSStringFromClass(oC);

    if ([originalClassName hasPrefix: @"WOScriptedClass"])
        originalClassName = NSStringFromClass([oC superclass]);

    if ([theObject isKindOfClass: NSClassFromString(@"WOComponent")]) {
        NSString *oN;
        oN = [(NSObject *)theObject name];
        return [NSString stringWithFormat: @"%@ (Name: %@)",
            originalClassName, oN];
    }

    if ([theObject isKindOfClass: NSClassFromString(@"WOApplication")]){
        NSString *oN = [(NSObject *)theObject name];
        return [NSString stringWithFormat: @"%@ (Name: %@)",
            originalClassName, oN];
    }

    if ([theObject isKindOfClass: NSClassFromString(@"WOSession")]){
        NSString *oN = [(NSObject *) [theObject application] name];
        return [NSString stringWithFormat: @"%@ (App: %@)",
            originalClassName, oN];
    }

    return originalClassName;
}

#define EMPTY_STRING @"[]"
+ (NSString *) methodLabelForInstance: (id) anObject
                          andSelector: (SEL) aSelector
/*"
 * Returns a string describing aSelector in terms of anObject.  The
 * returned strings indicates whether aSelector is a class or instance
 * method depending on whether anObject is a class object or instance of
 * some class.  If either anObject or aSelector is nil, the returned string
 * will indicate this.
"*/
{
    if(!anObject && !aSelector)
        return EMPTY_STRING;
    else if (anObject && aSelector) {
        NSString *className  = [self canonicalNameForClass:anObject];
        return [NSString stringWithFormat: @"[%@ %s%@]",
            className,
(anObject == NSClassFromString(className)) ? "+" : "-",
            NSStringFromSelector(aSelector)];
    } else if (anObject) {
        NSString *className  = [self canonicalNameForClass:anObject];
        return [NSString stringWithFormat: @"[%@ %s]",
            className,
(anObject == NSClassFromString(className)) ? "+" : "-"];
    } else if (aSelector) {
        return [NSString stringWithFormat: @"[%@]",
            NSStringFromSelector(aSelector)];
    }

    // NOT_REACHED
    return nil;
}

/* ========================================================================
** Time control messages
** ========================================================================
*/
static NSDate *_timerBase = nil;
static NSString *_timerFormat = nil;

- (void) resetIntervalTimerBase;
/*"
 * Resets the timer to have a base time of now.
"*/
{
    CFRSetVariableToObject(_timerBase, [NSDate date]);
}

- (NSDate *) intervalTimerBase
/*"
 * Returns the NSDate object containing the last time the timer was reset.
 * Returns nil if the timer was never reset.
"*/
{
    return _timerBase;
}

- (NSTimeInterval) timeIntervalSinceIntervalTimerBase;
/*"
 * Returns the NSTimeInterval between the time base time and now.  Raises
 * an exception if the timer was never reset.
 * 
 * This method will always return a positive interval;  the amount of time
 * that has elapsed since the base time and the time at which the method
 * was invoked
"*/
{
    if (_timerBase == nil)
        [NSException raise: NSInternalInconsistencyException
                    format: @"Timer must be reset with -resetIntervalTimerBase before time intervals can be measured."];

    return (- [_timerBase timeIntervalSinceNow]);
}

- (NSString *) intervalTimerLoggingFormat;
/*"
 * Returns the format string used to convert time intervals into human
 * readable text for logging purposes.
"*/
{
    return _timerFormat;
}

- (void) setIntervalTimerLoggingFormat: (NSString *) aFormat;
/*"
 * Sets the format string used to convert an NSTimeInterval into a human
 * readable string.  The default is @" (%.4f secs) ".
"*/
{
    if (aFormat == nil)
        aFormat = @"(%.4f secs) ";
    CFRSetVariableToObject(_timerFormat, aFormat);
}

- (NSString *) stringFromTimerIntervalSinceBase
/*"
 * Returns an NSString containing the elapased interval since the timer
 * base formatted using the interval timer logging format string.
"*/
{
    if (_timerFormat == nil)
        [self setIntervalTimerLoggingFormat: nil];

    return [NSString stringWithFormat: _timerFormat, [self timeIntervalSinceIntervalTimerBase]]; 
}

- (void) logTimeIntervalForMethod: (SEL) aMethod withFormat: (NSString *) aFormat, ...;
/*"
 * Generates a log message like #{-logMethod:withFormat:}, but the message
 * includes the amount of time that has elapsed since the timer base.  If
 * the timer has never been reset, this method will invoke #{-resetIntervalTimerBase}.
"*/
{
    NSString *timeFormatString;
    va_list ap;
    
    if (_timerBase == nil)
        [self resetIntervalTimerBase];

    if (aFormat != nil)
        timeFormatString = [[self stringFromTimerIntervalSinceBase] stringByAppendingString: aFormat];
    else
        timeFormatString = [self stringFromTimerIntervalSinceBase];

    va_start(ap, aFormat);

    //    - (void)logMethod: (SEL) aMethod withFormat:(NSString *)aFormat arguments:(va_list)argList;
    [self logMethod: aMethod withFormat: timeFormatString arguments: ap];

    va_end(ap);
}
@end

