/*
   Original Author: bbum

   Copyright 1999 by CodeFab, Inc.   All rights reserved.

   This notice may not be removed from this source code.
   The use and distribution of this software is governed by the
   terms of the MiscKit license agreement.  Refer to the license
   document included with the MiscKit distribution for the terms.
   Details of the license agreement may be found at
   www.misckit.com or within the MiscKit distribution.

   There may be more information about this software on
   www.codefab.com.  Send queries to ask_codefab@codefab.com.
*/

#import <Foundation/Foundation.h>
#import <EOControl/EOControl.h>
#import <EOAccess/EOAccess.h>
#import <WebObjects/WebObjects.h>

#import "NMSScope.h"
#import "NMSDebugger.h"
#import "NMSScriptFlavor.h"
#import "NMSScriptedObject.h"

#import "CodeFabSupport.h"

#import "Mssh.h"

#import "MsshController.h"

NSString *MSSHSScriptableAPIValueKey = @"MSSH Scriptable API Value Key";

static NSBundle *msshFrameworkBundle;

@interface NSObject (ScriptedAPI)
- (NSNumber *) main: (NSArray *) argumentArray;
@end

@implementation MsshController

+ (void) initialize
{
    NSAutoreleasePool *p = [[NSAutoreleasePool alloc] init];

    msshFrameworkBundle = [[NSBundle bundleForClass: self] retain];

    [p release];
}

+ (NSBundle *) msshFrameworkBundle;
{
    return msshFrameworkBundle;
}

+ classFromString: (NSString *) aClassName;
{
    return NSClassFromString(aClassName);
}

+ (void) _addGlobalsToWebScript
{
    NMSScope *globalScope = [[NSClassFromString(@"NMSScriptFlavor") scriptFlavorForName: @"WebScript"] theGlobalScope];

    /* DO / Run Loop stuff */
    [globalScope declareKey: @"NSConnectionReplyMode" type: @"id" initialValue: NSConnectionReplyMode];
    [globalScope declareKey: @"NSDefaultRunLoopMode" type: @"id" initialValue: NSDefaultRunLoopMode];
    [globalScope declareKey: @"NSConnectionDidDieNotification" type: @"id" initialValue: NSConnectionDidDieNotification];

    /* Distributed Notification Center stuff */
    [globalScope declareKey: @"NSLocalNotificationCenterType" type: @"id" initialValue: NSLocalNotificationCenterType];
    [globalScope declareKey: @"NSNotificationSuspensionBehaviorDrop" type: @"id" initialValue: [NSNumber numberWithInt: NSNotificationSuspensionBehaviorDrop]];
    [globalScope declareKey: @"NSNotificationSuspensionBehaviorCoalesce" type: @"id" initialValue: [NSNumber numberWithInt: NSNotificationSuspensionBehaviorCoalesce]];
    [globalScope declareKey: @"NSNotificationSuspensionBehaviorHold" type: @"id" initialValue: [NSNumber numberWithInt: NSNotificationSuspensionBehaviorHold]];
    [globalScope declareKey: @"NSNotificationSuspensionBehaviorDeliverImmediately" type: @"id" initialValue: [NSNumber numberWithInt: NSNotificationSuspensionBehaviorDeliverImmediately]];
    
    /* NSString's search modes */
    [globalScope declareKey: @"NSCaseInsensitiveSearch" type: @"id" initialValue: [NSNumber numberWithInt: NSCaseInsensitiveSearch]];
    [globalScope declareKey: @"NSLiteralSearch" type: @"id" initialValue: [NSNumber numberWithInt: NSLiteralSearch]];
    [globalScope declareKey: @"NSBackwardsSearch" type: @"id" initialValue: [NSNumber numberWithInt: NSBackwardsSearch]];
    [globalScope declareKey: @"NSAnchoredSearch" type: @"id" initialValue: [NSNumber numberWithInt: NSAnchoredSearch]];

    /* NSString encodings */
    [globalScope declareKey: @"NSASCIIStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSASCIIStringEncoding]];
    [globalScope declareKey: @"NSNEXTSTEPStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSNEXTSTEPStringEncoding]];
    [globalScope declareKey: @"NSJapaneseEUCStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSJapaneseEUCStringEncoding]];
    [globalScope declareKey: @"NSUTF8StringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSUTF8StringEncoding]];
    [globalScope declareKey: @"NSISOLatin1StringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSISOLatin1StringEncoding]];
    [globalScope declareKey: @"NSSymbolStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSSymbolStringEncoding]];
    [globalScope declareKey: @"NSNonLossyASCIIStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSNonLossyASCIIStringEncoding]];
    [globalScope declareKey: @"NSShiftJISStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSShiftJISStringEncoding]];
    [globalScope declareKey: @"NSISOLatin2StringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSISOLatin2StringEncoding]];
    [globalScope declareKey: @"NSUnicodeStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSUnicodeStringEncoding]];
    [globalScope declareKey: @"NSWindowsCP1251StringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSWindowsCP1251StringEncoding]];
    [globalScope declareKey: @"NSWindowsCP1252StringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSWindowsCP1252StringEncoding]];
    [globalScope declareKey: @"NSWindowsCP1253StringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSWindowsCP1253StringEncoding]];
    [globalScope declareKey: @"NSWindowsCP1250StringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSWindowsCP1250StringEncoding]];
    [globalScope declareKey: @"NSISO2022JPStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSISO2022JPStringEncoding]];
    [globalScope declareKey: @"NSMacOSRomanStringEncoding" type: @"id" initialValue: [NSNumber numberWithInt: NSMacOSRomanStringEncoding]];

    /* comparison results -- NSComparisonResult */
    [globalScope declareKey: @"NSOrderedAscending" type: @"id" initialValue: [NSNumber numberWithInt: NSOrderedAscending]];
    [globalScope declareKey: @"NSOrderedSame" type: @"id" initialValue: [NSNumber numberWithInt: NSOrderedSame]];
    [globalScope declareKey: @"NSOrderedDescending" type: @"id" initialValue: [NSNumber numberWithInt: NSOrderedDescending]];

    /* NSProcessInfo's OS identifiers */
    [globalScope declareKey: @"NSWindowsNTOperatingSystem" type: @"id" initialValue: [NSNumber numberWithInt: NSWindowsNTOperatingSystem]];
    [globalScope declareKey: @"NSWindows95OperatingSystem" type: @"id" initialValue: [NSNumber numberWithInt: NSWindows95OperatingSystem]];
    [globalScope declareKey: @"NSSolarisOperatingSystem" type: @"id" initialValue: [NSNumber numberWithInt: NSSolarisOperatingSystem]];
    [globalScope declareKey: @"NSHPUXOperatingSystem" type: @"id" initialValue: [NSNumber numberWithInt: NSHPUXOperatingSystem]];
    [globalScope declareKey: @"NSMACHOperatingSystem" type: @"id" initialValue: [NSNumber numberWithInt: NSMACHOperatingSystem]];
    [globalScope declareKey: @"NSSunOSOperatingSystem" type: @"id" initialValue: [NSNumber numberWithInt: NSSunOSOperatingSystem]];
    [globalScope declareKey: @"NSOSF1OperatingSystem" type: @"id" initialValue: [NSNumber numberWithInt: NSSunOSOperatingSystem]];

    /* Exceptions */
    [globalScope declareKey: @"NSGenericException" type: @"id" initialValue: NSGenericException];
    [globalScope declareKey: @"NSRangeException" type: @"id" initialValue: NSRangeException];
    [globalScope declareKey: @"NSInvalidArgumentException" type: @"id" initialValue: NSInvalidArgumentException];
    [globalScope declareKey: @"NSInternalInconsistencyException" type: @"id" initialValue: NSInternalInconsistencyException];
    [globalScope declareKey: @"NSDecimalNumberExactnessException" type: @"id" initialValue: NSDecimalNumberExactnessException];
    [globalScope declareKey: @"NSDecimalNumberOverflowException" type: @"id" initialValue: NSDecimalNumberOverflowException];
    [globalScope declareKey: @"NSDecimalNumberUnderflowException" type: @"id" initialValue: NSDecimalNumberUnderflowException];
    [globalScope declareKey: @"NSDecimalNumberDivideByZeroException" type: @"id" initialValue: NSDecimalNumberDivideByZeroException];
    [globalScope declareKey: @"NSParseErrorException" type: @"id" initialValue: NSParseErrorException];
    [globalScope declareKey: @"EOGeneralAdaptorException" type: @"id" initialValue: EOGeneralAdaptorException];
    [globalScope declareKey: @"EOGeneralDatabaseException" type: @"id" initialValue: EOGeneralDatabaseException];
    [globalScope declareKey: @"EOMoreThanOneException" type: @"id" initialValue: EOMoreThanOneException];
    [globalScope declareKey: @"EOValidationException" type: @"id" initialValue: EOValidationException];
    [globalScope declareKey: @"EOUnknownKeyException" type: @"id" initialValue: EOUnknownKeyException];
    [globalScope declareKey: @"EOQualifierVariableSubstitutionException" type: @"id" initialValue: EOQualifierVariableSubstitutionException];

    /* EO Delete Rules */
    [globalScope declareKey: @"EODeleteRuleNullify" type: @"id" initialValue: [NSNumber numberWithInt: EODeleteRuleNullify]];
    [globalScope declareKey: @"EODeleteRuleCascade" type: @"id" initialValue: [NSNumber numberWithInt: EODeleteRuleCascade]];
    [globalScope declareKey: @"EODeleteRuleDeny" type: @"id" initialValue: [NSNumber numberWithInt: EODeleteRuleDeny]];
    [globalScope declareKey: @"EODeleteRuleNoAction" type: @"id" initialValue: [NSNumber numberWithInt: EODeleteRuleNoAction]];

    /* EO Update Strategies */
    [globalScope declareKey: @"EOUpdateWithOptimisticLocking" type: @"id" initialValue: [NSNumber numberWithInt: EOUpdateWithOptimisticLocking]];
    [globalScope declareKey: @"EOUpdateWithPessimisticLocking" type: @"id" initialValue: [NSNumber numberWithInt: EOUpdateWithPessimisticLocking]];
    [globalScope declareKey: @"EOUpdateWithNoLocking" type: @"id" initialValue: [NSNumber numberWithInt: EOUpdateWithNoLocking]];

    /* EO Adaptor Operations */
    [globalScope declareKey: @"EOAdaptorLockOperator" type: @"id" initialValue: [NSNumber numberWithInt: EOAdaptorLockOperator]];
    [globalScope declareKey: @"EOAdaptorInsertOperator" type: @"id" initialValue: [NSNumber numberWithInt: EOAdaptorInsertOperator]];
    [globalScope declareKey: @"EOAdaptorUpdateOperator" type: @"id" initialValue: [NSNumber numberWithInt: EOAdaptorUpdateOperator]];
    [globalScope declareKey: @"EOAdaptorDeleteOperator" type: @"id" initialValue: [NSNumber numberWithInt: EOAdaptorDeleteOperator]];
    [globalScope declareKey: @"EOAdaptorStoredProcedureOperator" type: @"id" initialValue: [NSNumber numberWithInt: EOAdaptorStoredProcedureOperator]];

    /* EO Join Semantic */
    [globalScope declareKey: @"EOInnerJoin" type: @"id" initialValue: [NSNumber numberWithInt: EOInnerJoin]];
    [globalScope declareKey: @"EOFullOuterJoin" type: @"id" initialValue: [NSNumber numberWithInt: EOFullOuterJoin]];
    [globalScope declareKey: @"EOLeftOuterJoin" type: @"id" initialValue: [NSNumber numberWithInt: EOLeftOuterJoin]];
    [globalScope declareKey: @"EORightOuterJoin" type: @"id" initialValue: [NSNumber numberWithInt: EORightOuterJoin]];

    /* EO Binding Dictionary Keys */
    [globalScope declareKey: @"EOBindVariableNameKey" type: @"id" initialValue: EOBindVariableNameKey];
    [globalScope declareKey: @"EOBindVariableAttributeKey" type: @"id" initialValue: EOBindVariableAttributeKey];
    [globalScope declareKey: @"EOBindVariableValueKey" type: @"id" initialValue: EOBindVariableValueKey];
    [globalScope declareKey: @"EOBindVariablePlaceHolderKey" type: @"id" initialValue: EOBindVariablePlaceHolderKey];
    [globalScope declareKey: @"EOBindVariableColumnKey" type: @"id" initialValue: EOBindVariableColumnKey];
    
    /* EO Notifications */
    [globalScope declareKey: @"EOGlobalIDChangedNotification" type: @"id" initialValue: EOGlobalIDChangedNotification];

    /* NSTask / NSFileHandle Notifications & Keys */
    [globalScope declareKey: @"NSFileHandleOperationException" type: @"id" initialValue: NSFileHandleOperationException];
    [globalScope declareKey: @"NSFileHandleReadCompletionNotification" type: @"id" initialValue: NSFileHandleReadCompletionNotification];
    [globalScope declareKey: @"NSFileHandleReadToEndOfFileCompletionNotification" type: @"id" initialValue: NSFileHandleReadToEndOfFileCompletionNotification];
    [globalScope declareKey: @"NSFileHandleConnectionAcceptedNotification" type: @"id" initialValue: NSFileHandleConnectionAcceptedNotification];
    [globalScope declareKey: @"NSFileHandleDataAvailableNotification" type: @"id" initialValue: NSFileHandleDataAvailableNotification];
    [globalScope declareKey: @"NSFileHandleNotificationDataItem" type: @"id" initialValue: NSFileHandleNotificationDataItem];
    [globalScope declareKey: @"NSFileHandleNotificationFileHandleItem" type: @"id" initialValue: NSFileHandleNotificationFileHandleItem];
    [globalScope declareKey: @"NSFileHandleNotificationMonitorModes" type: @"id" initialValue: NSFileHandleNotificationMonitorModes];
    [globalScope declareKey: @"NSTaskDidTerminateNotification" type: @"id" initialValue: NSTaskDidTerminateNotification];

    /* Scriptable API Support */
    [globalScope declareKey: @"MSSHScannerValueKey" type: @"id" initialValue: MSSHScannerValueKey];
    [globalScope declareKey: @"MSSHSScriptableAPIValueKey" type: @"id" initialValue: MSSHSScriptableAPIValueKey];

    /* NSFileManager Support */
    [globalScope declareKey: @"NSFileType" type: @"id" initialValue: NSFileType];
    [globalScope declareKey: @"NSFileTypeDirectory" type: @"id" initialValue: NSFileTypeDirectory];
    [globalScope declareKey: @"NSFileTypeRegular" type: @"id" initialValue: NSFileTypeRegular];
    [globalScope declareKey: @"NSFileTypeSymbolicLink" type: @"id" initialValue: NSFileTypeSymbolicLink];
    [globalScope declareKey: @"NSFileTypeSocket" type: @"id" initialValue: NSFileTypeSocket];
    [globalScope declareKey: @"NSFileTypeCharacterSpecial" type: @"id" initialValue: NSFileTypeCharacterSpecial];
    [globalScope declareKey: @"NSFileTypeBlockSpecial" type: @"id" initialValue: NSFileTypeBlockSpecial];
    [globalScope declareKey: @"NSFileTypeUnknown" type: @"id" initialValue: NSFileTypeUnknown];
    [globalScope declareKey: @"NSFileSize" type: @"id" initialValue: NSFileSize];
    [globalScope declareKey: @"NSFileModificationDate" type: @"id" initialValue: NSFileModificationDate];
    [globalScope declareKey: @"NSFileReferenceCount" type: @"id" initialValue: NSFileReferenceCount];
    [globalScope declareKey: @"NSFileDeviceIdentifier" type: @"id" initialValue: NSFileDeviceIdentifier];
    [globalScope declareKey: @"NSFileOwnerAccountName" type: @"id" initialValue: NSFileOwnerAccountName];
    [globalScope declareKey: @"NSFileGroupOwnerAccountName" type: @"id" initialValue: NSFileGroupOwnerAccountName];
    [globalScope declareKey: @"NSFilePosixPermissions" type: @"id" initialValue: NSFilePosixPermissions];
    [globalScope declareKey: @"NSFileSystemNumber" type: @"id" initialValue: NSFileSystemNumber];
    [globalScope declareKey: @"NSFileSystemFileNumber" type: @"id" initialValue: NSFileSystemFileNumber];

    [globalScope declareKey: @"NSFileSystemSize" type: @"id" initialValue: NSFileSystemSize];
    [globalScope declareKey: @"NSFileSystemFreeSize" type: @"id" initialValue: NSFileSystemFreeSize];
    [globalScope declareKey: @"NSFileSystemNodes" type: @"id" initialValue: NSFileSystemFileNumber];
    [globalScope declareKey: @"NSFileSystemFreeNodes" type: @"id" initialValue: NSFileSystemFreeNodes];
}

+ (void) initializeWebScriptInterpreter;
{
    [self _addGlobalsToWebScript];
}

typedef enum {
    MSSHByExtension,
    MSSHAsClass,
    MSSHAsObject,
    MSSHAsBundle
} MSSHScriptParsingMode;	

#define USAGE @"mssh [options] <<path> [arg0 [arg1 [arg2 [...]]]]\n\t{-object,-class,-bundle} Load path as class, object(default) or bundle\n\t-trace <<Aspect>>  Enabled tracing of aspect.\n\t-trace EOAdaptor	Enable EOAdaptor tracing\n\t-trace NSException	Turn on exception tracing."

static NSMutableArray *scriptArguments = nil;

+ (NSArray *) scriptArguments;
{
    return scriptArguments;
}

+ (int) main
{
NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
    Class nmsScriptedObjectClass = NSClassFromString(@"NMSScriptedObject");
    Class nmsScriptFlavorClass = NSClassFromString(@"NMSScriptFlavor");
    NSArray *unprocessedArray;
    NSProcessInfo *processInfo;
    NSRange argRange;
    int exitCode = -254;
    NSNumber *exitCodeNumber;
    MSSHScriptParsingMode parseMode;
    BOOL parsingModeAlreadySet = NO;
    BOOL encounteredOne = NO;
    id scriptFlavor = nil;

    processInfo = [NSProcessInfo processInfo];
    unprocessedArray = [processInfo arguments];

    if ([unprocessedArray count] < 2) {
        NSLog(USAGE);
        return exitCode;
    }

    argRange.location = 1;
    argRange.length = [unprocessedArray count] - 1;

    scriptArguments = [NSMutableArray arrayWithArray: [unprocessedArray subarrayWithRange: argRange]];

    parseMode = MSSHByExtension;

    do {
        if ([scriptArguments containsObject: @"-object"] == YES) {
            parseMode = MSSHAsObject;
            if (parsingModeAlreadySet == YES) {
                NSLog(@"Only one of -object, -bundle or -class should be specified.\n%@", USAGE);
                return exitCode;
            }
            parsingModeAlreadySet = YES;
            [scriptArguments removeObject: @"-object"];
            encounteredOne = YES;
            scriptFlavor = [nmsScriptFlavorClass scriptFlavorForName: @"WebScript"];
        } else if ([scriptArguments containsObject: @"-class"] == YES) {
            parseMode = MSSHAsClass;
            if (parsingModeAlreadySet == YES) {
                NSLog(@"Only one of -object, -bundle or -class should be specified.\n%@", USAGE);
                return exitCode;
            }
            parsingModeAlreadySet = YES;
            [scriptArguments removeObject: @"-class"];
            encounteredOne = YES;
            scriptFlavor = [nmsScriptFlavorClass scriptFlavorForName: @"WebScript"];
        } else if ([scriptArguments containsObject: @"-bundle"] == YES) {
            parseMode = MSSHAsBundle;
            if (parsingModeAlreadySet == YES) {
                NSLog(@"Only one of -object, -bundle or -class should be specified.\n%@", USAGE);
                return exitCode;
            }
            parsingModeAlreadySet = YES;
            [scriptArguments removeObject: @"-bundle"];
            encounteredOne = YES;
        } else if ([scriptArguments containsObject: @"-preload"] == YES) {
            int i = [scriptArguments indexOfObject: @"-preload"];
           NSString *bundlePath;
           [scriptArguments removeObject: @"-preload"];
           bundlePath = [scriptArguments objectAtIndex: i];
           [scriptArguments removeObjectAtIndex: i];
           [[NSBundle bundleWithPath: bundlePath] principalClass];
           encounteredOne = YES;
       } else
            encounteredOne = NO;
    } while (encounteredOne == YES);

    while ([scriptArguments containsObject: @"-trace"] == YES) {
        unsigned int aspectIndex = [scriptArguments indexOfObject: @"-trace"];
        NSString *aspect;

        NS_DURING
            [scriptArguments removeObjectAtIndex: aspectIndex];
            aspect = [scriptArguments objectAtIndex: aspectIndex];
            if ([aspect isEqualToString: @"EOAdapator"] == YES) {
                [NSObject logWithFormat: @"EOAdaptor debugging enabled."];
                [NSClassFromString(@"EOAdaptorContext") setDebugEnabledDefault: YES];
            } else if ([aspect isEqualToString: @"NSException"] == YES) {
                [CFRExceptionTracer installExceptionTracing];
                [CFRExceptionTracer setExceptionTracingEnabled: YES];
                } else {
                    [[NSClassFromString(@"NMSDebugger") defaultDebugger] startTracingAspect: aspect];
                }
            [scriptArguments removeObjectAtIndex: aspectIndex];
        NS_HANDLER
            NSLog(USAGE);
            NSLog(@"Exception (likely because the arguments to mssh were wrong):\n\t%@", localException);
        NS_ENDHANDLER
    }

    NS_DURING
        NSString *scriptPath = [scriptArguments objectAtIndex: 0];
        NSMutableString *scriptString;
        NSObject *entryPointObject;
        Class entryPointClass = Nil;
        NSString *pathExtension = [[scriptPath pathExtension] lowercaseString];

        if (parseMode == MSSHByExtension) {
            if ([pathExtension isEqualToString: @"wos"] == YES) {
                parseMode = MSSHAsObject;
                scriptFlavor = [nmsScriptFlavorClass scriptFlavorForName: @"WebScript"];
            } else if ([pathExtension isEqualToString: @"wosc"] == YES) {
                parseMode = MSSHAsClass;
                scriptFlavor = [nmsScriptFlavorClass scriptFlavorForName: @"WebScript"];
            } else if ([pathExtension isEqualToString: @"bundle"] == YES)
                parseMode = MSSHAsBundle;
        }

        if (parseMode == MSSHAsBundle) {
            NSBundle *theBundle = [NSBundle bundleWithPath: scriptPath];
            NSArray *theClasses;
            theClasses = [nmsScriptedObjectClass scriptedClassesForBundle:theBundle inDirectory:nil];

            entryPointClass = [theBundle principalClass];
        }

        if (entryPointClass == Nil) { // not a bundle-- no entry point defined
            scriptString = [NSMutableString stringWithContentsOfFile: scriptPath];


            if ( (scriptString == nil) || ([scriptString length] == 0) ) {
                NSLog(@"Couldn't find or read script file %@.\n%@", scriptPath, USAGE);
                return exitCode;
            }
            
            if ([scriptString hasPrefix: @"#!"] == YES) {
                NSRange slashBangRange = {0,2};
                [scriptString replaceCharactersInRange: slashBangRange withString: @"//"];
            }
        }

        [self _addGlobalsToWebScript];

        if (parseMode == MSSHAsObject) {
            if (scriptFlavor == nil)
                scriptFlavor = [nmsScriptFlavorClass scriptFlavorForPath: scriptPath];

            entryPointObject = [nmsScriptedObjectClass scriptedObjectForString: scriptString scriptFlavor: scriptFlavor];
            exitCodeNumber = [(NMSScriptedObject *)entryPointObject evaluateMethodNamed: @"main:" count:1 arguments: &scriptArguments];
        } else if ((parseMode == MSSHAsClass) || (entryPointClass != Nil)) {
            if (entryPointClass == Nil) {
                if (scriptFlavor == nil)
                    scriptFlavor = [nmsScriptFlavorClass scriptFlavorForPath: scriptPath];

                entryPointClass = [nmsScriptedObjectClass scriptedClassForString: scriptString scriptFlavor: scriptFlavor];
            }

            if ([entryPointClass respondsToSelector: @selector(main:)] == YES)
                exitCodeNumber = [entryPointClass main: scriptArguments];
            else {
                entryPointObject = [[[entryPointClass alloc] init] autorelease];
                exitCodeNumber = [entryPointObject main: scriptArguments];
            }
        }

        exitCode = [exitCodeNumber shortValue];
    NS_HANDLER
        NSLog(@"Exception: %@", localException);
    NS_ENDHANDLER

    [pool release];

    exit(exitCode);       // insure the process exit status is 0
    return exitCode;      // ...and make main fit the ANSI spec.
}

+ (void) importScriptedMethodsForPath:(NSString *)aPath;
{
    static int	categoryCounter = 1;
    Class		nmsScriptedObjectClass;
    Class		nmsScriptFlavorClass;
    id			scriptFlavor = nil;
    NSString	*scriptString;

    nmsScriptedObjectClass = NSClassFromString(@"NMSScriptedObject");
    nmsScriptFlavorClass = NSClassFromString(@"NMSScriptFlavor");
    scriptFlavor = [nmsScriptFlavorClass scriptFlavorForPath: aPath];
    scriptString = [NSString stringWithContentsOfFile: aPath];
    if (![scriptString rangeOfString:@"@implementation "].length)
    {
        scriptString = [NSString stringWithFormat:@"%@%i%@%@%@",
            @"@implementation NMSScriptedObject (ImportScriptWithPath",
            categoryCounter++,
            @");	",
            scriptString,
            @"\n@end"];
    }
    [nmsScriptedObjectClass scriptedClassForString:scriptString
                                      scriptFlavor:scriptFlavor
                                    usingClassName:@""
                                    superClassName:@"NSObject"];
}

+ (Class) importScriptedClassForPath: (NSString *) aPath;
{
    return [NSClassFromString(@"NMSScriptedObject") scriptedClassForPath:aPath encoding:NSNEXTSTEPStringEncoding];
}

static NSMutableDictionary *transientMutableDictionary;
+ (void) setTransientValue: anObject forName: (NSString *) aName;
{
    if (transientMutableDictionary == nil)
        transientMutableDictionary = [[NSMutableDictionary dictionary] retain];

    if (anObject == nil)
        [transientMutableDictionary removeObjectForKey: aName];
    else
        [transientMutableDictionary setObject: anObject forKey: aName];
}

+ transientValueForName: (NSString *) aName;
{
    return [transientMutableDictionary objectForKey: aName];
}

+ (void) setValue: anObjectValue forGlobalVariable: (NSString *) aGlobalVariableName;
{
    NMSScope *globalScope = [[NSClassFromString(@"NMSScriptFlavor") scriptFlavorForName: @"WebScript"] theGlobalScope];

    /* NSString's search modes */
    [globalScope declareKey: aGlobalVariableName type: @"id" initialValue: anObjectValue];
}

+ valueForGlobalVariable: (NSString *) aGlobalVariableName;
{
    NMSScope *globalScope = [[NSClassFromString(@"NMSScriptFlavor") scriptFlavorForName: @"WebScript"] theGlobalScope];

    return [globalScope valueForKey: aGlobalVariableName];
}
@end
